from .registry import Registry, RegistryType
from .registry_extension import REGISTRY_EXTENSION_TYPE, RegistryExtension

__all__ = [
    "Registry",
    "RegistryType",
    "REGISTRY_EXTENSION_TYPE",
    "RegistryExtension",
]



from omu.errors import InvalidPacket
from omu.identifier import Identifier
from omu.serializer import Serializable

from .packet import Packet, PacketData, PacketType


class PacketMapper(Serializable[Packet, PacketData]):
    def __init__(self) -> None:
        self._map: dict[Identifier, PacketType] = {}

    def register(self, *packet_types: PacketType) -> None:
        for packet_type in packet_types:
            if self._map.get(packet_type.id):
                raise ValueError(f"Packet id {packet_type.id} already registered")
            self._map[packet_type.id] = packet_type

    def serialize(self, item: Packet) -> PacketData:
        return PacketData(
            type=item.type.id.key(),
            data=item.type.serializer.serialize(item.data),
        )

    def deserialize(self, item: PacketData) -> Packet:
        id = Identifier.from_key(item.type)
        packet_type = self._map.get(id)
        if not packet_type:
            raise InvalidPacket(id, f"Packet type {id} not registered")
        try:
            data = packet_type.serializer.deserialize(item.data)
        except Exception as e:
            raise InvalidPacket(id, "Failed to deserialize packet data") from e
        return Packet(
            type=packet_type,
            data=data,
        )



from __future__ import annotations

import asyncio

from omu.address import Address
from omu.app import App
from omu.event_emitter import Unlisten
from omu.extension import ExtensionRegistry
from omu.extension.asset import (
    ASSET_EXTENSION_TYPE,
    AssetExtension,
)
from omu.extension.dashboard import (
    DASHBOARD_EXTENSION_TYPE,
    DashboardExtension,
)
from omu.extension.endpoint import (
    ENDPOINT_EXTENSION_TYPE,
    EndpointExtension,
)
from omu.extension.i18n import (
    I18N_EXTENSION_TYPE,
    I18nExtension,
)
from omu.extension.logger import (
    LOGGER_EXTENSION_TYPE,
    LoggerExtension,
)
from omu.extension.permission import (
    PERMISSION_EXTENSION_TYPE,
    PermissionExtension,
)
from omu.extension.plugin import (
    PLUGIN_EXTENSION_TYPE,
    PluginExtension,
)
from omu.extension.registry import (
    REGISTRY_EXTENSION_TYPE,
    RegistryExtension,
)
from omu.extension.server import (
    SERVER_EXTENSION_TYPE,
    ServerExtension,
)
from omu.extension.signal import (
    SIGNAL_EXTENSION_TYPE,
    SignalExtension,
)
from omu.extension.table import (
    TABLE_EXTENSION_TYPE,
    TableExtension,
)
from omu.helper import Coro, asyncio_error_logger
from omu.network import Network
from omu.network.packet import Packet, PacketType
from omu.network.websocket_connection import WebsocketsConnection
from omu.token import JsonTokenProvider, TokenProvider

from .client import Client, ClientEvents


class Omu(Client):
    def __init__(
        self,
        app: App,
        address: Address | None = None,
        token: TokenProvider | None = None,
        connection: WebsocketsConnection | None = None,
        extension_registry: ExtensionRegistry | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        self._loop = self.ensure_loop(loop)
        self._ready = False
        self._running = False
        self._event = ClientEvents()
        self._app = app
        self.address = address or Address("127.0.0.1", 26423)
        self._network = Network(
            self,
            self.address,
            token or JsonTokenProvider(),
            connection or WebsocketsConnection(self, self.address),
        )
        self._extensions = extension_registry or ExtensionRegistry(self)

        self._endpoints = self.extensions.register(ENDPOINT_EXTENSION_TYPE)
        self._plugins = self.extensions.register(PLUGIN_EXTENSION_TYPE)
        self._tables = self.extensions.register(TABLE_EXTENSION_TYPE)
        self._registry = self.extensions.register(REGISTRY_EXTENSION_TYPE)
        self._signal = self.extensions.register(SIGNAL_EXTENSION_TYPE)
        self._permissions = self.extensions.register(PERMISSION_EXTENSION_TYPE)
        self._server = self.extensions.register(SERVER_EXTENSION_TYPE)
        self._assets = self.extensions.register(ASSET_EXTENSION_TYPE)
        self._dashboard = self.extensions.register(DASHBOARD_EXTENSION_TYPE)
        self._i18n = self.extensions.register(I18N_EXTENSION_TYPE)
        self._logger = self.extensions.register(LOGGER_EXTENSION_TYPE)

    def ensure_loop(
        self, loop: asyncio.AbstractEventLoop | None
    ) -> asyncio.AbstractEventLoop:
        if loop is None:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
        loop.set_exception_handler(asyncio_error_logger)
        return loop

    @property
    def ready(self) -> bool:
        return self._ready

    @property
    def app(self) -> App:
        return self._app

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def network(self) -> Network:
        return self._network

    @property
    def extensions(self) -> ExtensionRegistry:
        return self._extensions

    @property
    def endpoints(self) -> EndpointExtension:
        return self._endpoints

    @property
    def plugins(self) -> PluginExtension:
        return self._plugins

    @property
    def tables(self) -> TableExtension:
        return self._tables

    @property
    def registry(self) -> RegistryExtension:
        return self._registry

    @property
    def signal(self) -> SignalExtension:
        return self._signal

    @property
    def assets(self) -> AssetExtension:
        return self._assets

    @property
    def server(self) -> ServerExtension:
        return self._server

    @property
    def permissions(self) -> PermissionExtension:
        return self._permissions

    @property
    def dashboard(self) -> DashboardExtension:
        return self._dashboard

    @property
    def i18n(self) -> I18nExtension:
        return self._i18n

    @property
    def logger(self) -> LoggerExtension:
        return self._logger

    @property
    def running(self) -> bool:
        return self._running

    async def send[T](self, type: PacketType[T], data: T) -> None:
        await self._network.send(Packet(type, data))

    def run(
        self,
        *,
        loop: asyncio.AbstractEventLoop | None = None,
        reconnect: bool = True,
    ) -> None:
        self.loop = self.ensure_loop(loop)
        try:
            self.loop.set_exception_handler(asyncio_error_logger)
            self.loop.create_task(self.start(reconnect=reconnect))
            self.loop.run_forever()
        finally:
            self.loop.close()
            asyncio.run(self.stop())

    async def start(self, *, reconnect: bool = True) -> None:
        if self._running:
            raise RuntimeError("Already running")
        self._running = True
        self.loop.create_task(self._network.connect(reconnect=reconnect))
        await self._event.started()

    async def stop(self) -> None:
        if not self._running:
            raise RuntimeError("Not running")
        self._running = False
        await self._network.disconnect()
        await self._event.stopped()

    @property
    def event(self) -> ClientEvents:
        return self._event

    def on_ready(self, coro: Coro[[], None]) -> Unlisten:
        if self._ready:
            self.loop.create_task(coro())
        return self.event.ready.listen(coro)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True, slots=True)
class EndpointType[Req, Res]:
    id: Identifier
    request_serializer: Serializable[Req, bytes]
    response_serializer: Serializable[Res, bytes]
    permission_id: Identifier | None = None

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        request_serializer: Serializable[Req, Any] | None = None,
        response_serializer: Serializable[Res, Any] | None = None,
        permission_id: Identifier | None = None,
    ):
        request_serializer = Serializer.of(
            request_serializer or Serializer.noop()
        ).to_json()
        response_serializer = Serializer.of(
            response_serializer or Serializer.noop()
        ).to_json()
        return cls(
            id=identifier / name,
            request_serializer=request_serializer,
            response_serializer=response_serializer,
            permission_id=permission_id,
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        request_serializer: Serializable[Req, bytes],
        response_serializer: Serializable[Res, bytes],
        permission_id: Identifier | None = None,
    ):
        return cls(
            id=identifier / name,
            request_serializer=request_serializer,
            response_serializer=response_serializer,
            permission_id=permission_id,
        )



from .address import Address
from .app import App
from .identifier import Identifier
from .network import Network, NetworkStatus
from .omu import Omu
from .plugin import Plugin
from .version import VERSION

__version__ = VERSION
__all__ = [
    "Address",
    "Network",
    "NetworkStatus",
    "Omu",
    "App",
    "Identifier",
    "Plugin",
]



from dataclasses import dataclass

from omu.bytebuffer import ByteReader, ByteWriter
from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.identifier import Identifier
from omu.serializer import Serializer

ASSET_EXTENSION_TYPE = ExtensionType(
    "asset",
    lambda client: AssetExtension(client),
    lambda: [],
)


@dataclass(frozen=True, slots=True)
class File:
    identifier: Identifier
    buffer: bytes


class FileSerializer:
    @classmethod
    def serialize(cls, item: File) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.identifier.key())
        writer.write_byte_array(item.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> File:
        with ByteReader(item) as reader:
            identifier = Identifier.from_key(reader.read_string())
            value = reader.read_byte_array()
        return File(identifier, value)


class FileArraySerializer:
    @classmethod
    def serialize(cls, item: list[File]) -> bytes:
        writer = ByteWriter()
        writer.write_int(len(item))
        for file in item:
            writer.write_string(file.identifier.key())
            writer.write_byte_array(file.buffer)
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> list[File]:
        with ByteReader(item) as reader:
            count = reader.read_int()
            files: list[File] = []
            for _ in range(count):
                identifier = Identifier.from_key(reader.read_string())
                value = reader.read_byte_array()
                files.append(File(identifier, value))
        return files


ASSET_UPLOAD_PERMISSION_ID = ASSET_EXTENSION_TYPE / "upload"
ASSET_UPLOAD_ENDPOINT = EndpointType[File, Identifier].create_serialized(
    ASSET_EXTENSION_TYPE,
    "upload",
    request_serializer=FileSerializer,
    response_serializer=Serializer.model(Identifier).to_json(),
    permission_id=ASSET_UPLOAD_PERMISSION_ID,
)
ASSET_UPLOAD_MANY_ENDPOINT = EndpointType[
    list[File], list[Identifier]
].create_serialized(
    ASSET_EXTENSION_TYPE,
    "upload_many",
    request_serializer=FileArraySerializer,
    response_serializer=Serializer.model(Identifier).to_array().to_json(),
    permission_id=ASSET_UPLOAD_PERMISSION_ID,
)
ASSET_DOWNLOAD_PERMISSION_ID = ASSET_EXTENSION_TYPE / "download"
ASSET_DOWNLOAD_ENDPOINT = EndpointType[Identifier, File].create_serialized(
    ASSET_EXTENSION_TYPE,
    "download",
    request_serializer=Serializer.model(Identifier).to_json(),
    response_serializer=FileSerializer,
    permission_id=ASSET_DOWNLOAD_PERMISSION_ID,
)
ASSET_DOWNLOAD_MANY_ENDPOINT = EndpointType[
    list[Identifier], list[File]
].create_serialized(
    ASSET_EXTENSION_TYPE,
    "download_many",
    request_serializer=Serializer.model(Identifier).to_array().to_json(),
    response_serializer=FileArraySerializer,
    permission_id=ASSET_DOWNLOAD_PERMISSION_ID,
)


class AssetExtension(Extension):
    def __init__(self, client: Client) -> None:
        self.client = client

    async def upload(self, file: File) -> Identifier:
        return await self.client.endpoints.call(ASSET_UPLOAD_ENDPOINT, file)

    async def upload_many(self, files: list[File]) -> list[Identifier]:
        return await self.client.endpoints.call(ASSET_UPLOAD_MANY_ENDPOINT, files)

    async def download(self, identifier: Identifier) -> File:
        return await self.client.endpoints.call(ASSET_DOWNLOAD_ENDPOINT, identifier)

    async def download_many(self, identifiers: list[Identifier]) -> list[File]:
        return await self.client.endpoints.call(
            ASSET_DOWNLOAD_MANY_ENDPOINT, identifiers
        )

    def url(self, identifier: Identifier) -> str:
        address = self.client.network.address
        protocol = "https" if address.secure else "http"
        return f"{protocol}://{address.host}:{address.port}/asset?id={identifier.key()}"

    def proxy(self, url: str) -> str:
        address = self.client.network.address
        protocol = "https" if address.secure else "http"
        return f"{protocol}://{address.host}:{address.port}/proxy?url={url}"



from .dashboard_extension import (
    DASHBOARD_EXTENSION_TYPE,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_SET_PERMISSION_ID,
    DASHOBARD_APP_EDIT_PERMISSION_ID,
    DASHOBARD_APP_READ_PERMISSION_ID,
    DashboardExtension,
)
from .packets import PermissionRequestPacket

__all__ = [
    "DASHBOARD_EXTENSION_TYPE",
    "DASHBOARD_SET_PERMISSION_ID",
    "DASHBOARD_OPEN_APP_PERMISSION_ID",
    "DASHOBARD_APP_READ_PERMISSION_ID",
    "DASHOBARD_APP_EDIT_PERMISSION_ID",
    "DashboardExtension",
    "PermissionRequestPacket",
]



from omu.app import App
from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.extension.registry import RegistryType
from omu.extension.table import TablePermissions, TableType
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.serializer import Serializer

SERVER_EXTENSION_TYPE = ExtensionType(
    "server", lambda client: ServerExtension(client), lambda: []
)

SERVER_APPS_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "apps" / "read"
SERVER_APP_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "apps",
    App,
    permissions=TablePermissions(
        read=SERVER_APPS_READ_PERMISSION_ID,
    ),
)
SERVER_SESSIONS_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "sessions" / "read"
SERVER_SESSION_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "sessions",
    App,
    permissions=TablePermissions(
        read=SERVER_SESSIONS_READ_PERMISSION_ID,
    ),
)
SERVER_SHUTDOWN_PERMISSION_ID = SERVER_EXTENSION_TYPE / "shutdown"
SHUTDOWN_ENDPOINT_TYPE = EndpointType[bool, bool].create_json(
    SERVER_EXTENSION_TYPE,
    "shutdown",
    permission_id=SERVER_SHUTDOWN_PERMISSION_ID,
)
REQUIRE_APPS_PACKET_TYPE = PacketType[list[Identifier]].create_json(
    SERVER_EXTENSION_TYPE,
    "require_apps",
    serializer=Serializer.model(Identifier).to_array(),
)
VERSION_REGISTRY_TYPE = RegistryType[str | None].create_json(
    SERVER_EXTENSION_TYPE,
    "version",
    default_value=None,
)


class ServerExtension(Extension):
    def __init__(self, client: Client) -> None:
        self._client = client
        self.apps = client.tables.get(SERVER_APP_TABLE_TYPE)
        self.sessions = client.tables.get(SERVER_APP_TABLE_TYPE)
        self.required_apps: set[Identifier] = set()
        client.network.register_packet(REQUIRE_APPS_PACKET_TYPE)
        client.network.add_task(self.on_task)

    async def on_task(self) -> None:
        if self.required_apps:
            await self._client.send(REQUIRE_APPS_PACKET_TYPE, [*self.required_apps])

    async def shutdown(self, restart: bool = False) -> bool:
        return await self._client.endpoints.call(SHUTDOWN_ENDPOINT_TYPE, restart)

    def require(self, *app_ids: Identifier) -> None:
        if self._client.running:
            raise RuntimeError("Cannot require apps after the client has started")
        self.required_apps.update(app_ids)



from __future__ import annotations

import aiohttp
from aiohttp import web

from omu.address import Address
from omu.bytebuffer import ByteReader, ByteWriter
from omu.client import Client
from omu.serializer import Serializable

from .connection import Connection
from .packet import Packet, PacketData


class WebsocketsConnection(Connection):
    def __init__(self, client: Client, address: Address):
        self._client = client
        self._address = address
        self._connected = False
        self._socket: aiohttp.ClientWebSocketResponse | None = None
        self._session: aiohttp.ClientSession | None = None

    @property
    def _ws_endpoint(self) -> str:
        protocol = "wss" if self._address.secure else "ws"
        host = self._address.host or "127.0.0.1"
        port = self._address.port
        return f"{protocol}://{host}:{port}/ws"

    async def connect(self) -> None:
        if self._socket and not self._socket.closed:
            raise RuntimeError("Already connected")
        self._session = aiohttp.ClientSession()
        self._socket = await self._session.ws_connect(self._ws_endpoint)
        self._connected = True

    async def send(
        self, packet: Packet, packet_mapper: Serializable[Packet, PacketData]
    ) -> None:
        if not self._socket or self._socket.closed or not self._connected:
            raise RuntimeError("Not connected")
        packet_data = packet_mapper.serialize(packet)
        writer = ByteWriter()
        writer.write_string(packet_data.type)
        writer.write_byte_array(packet_data.data)
        await self._socket.send_bytes(writer.finish())

    async def receive(self, packet_mapper: Serializable[Packet, PacketData]) -> Packet:
        if not self._socket or self._socket.closed:
            raise RuntimeError("Not connected")
        msg = await self._socket.receive()
        if msg.type in {
            web.WSMsgType.CLOSE,
            web.WSMsgType.CLOSED,
            web.WSMsgType.CLOSING,
            web.WSMsgType.ERROR,
        }:
            raise RuntimeError(f"Socket {msg.type.name.lower()}")
        if msg.data is None:
            raise RuntimeError("Received empty message")
        if msg.type == web.WSMsgType.TEXT:
            raise RuntimeError("Received text message")
        elif msg.type == web.WSMsgType.BINARY:
            with ByteReader(msg.data) as reader:
                event_type = reader.read_string()
                event_data = reader.read_byte_array()
            packet_data = PacketData(event_type, event_data)
            return packet_mapper.deserialize(packet_data)
        else:
            raise RuntimeError(f"Unknown message type {msg.type}")

    async def close(self) -> None:
        if not self._socket or self._socket.closed:
            return
        if self._socket:
            try:
                await self._socket.close()
            except AttributeError:
                pass
        self._socket = None
        self._connected = False

    @property
    def closed(self) -> bool:
        return not self._socket or self._socket.closed

    def __del__(self):
        if not self.closed:
            self._client.loop.create_task(self.close())
        if self._session:
            self._client.loop.create_task(self._session.close())



from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.registry import RegistryType
from omu.extension.registry.packets import RegistryPermissions
from omu.localization import Locale, LocalizedText

I18N_EXTENSION_TYPE = ExtensionType(
    "i18n", lambda client: I18nExtension(client), lambda: []
)
I18N_SET_LOCALES_PERMISSION_ID = I18N_EXTENSION_TYPE / "locales" / "set"
I18N_GET_LOCALES_PERMISSION_ID = I18N_EXTENSION_TYPE / "locales" / "get"
I18N_LOCALES_REGISTRY_TYPE = RegistryType[list[Locale]].create_json(
    I18N_EXTENSION_TYPE,
    name="locales",
    default_value=[],
    permissions=RegistryPermissions(
        read=I18N_GET_LOCALES_PERMISSION_ID,
        write=I18N_SET_LOCALES_PERMISSION_ID,
    ),
)


class I18nExtension(Extension):
    def __init__(self, client: Client):
        self.client = client
        client.permissions.require(I18N_GET_LOCALES_PERMISSION_ID)
        self.locales_registry = client.registry.get(I18N_LOCALES_REGISTRY_TYPE)
        self.locales: list[Locale] = []

    def translate(self, localized_text: LocalizedText) -> str:
        if not self.locales:
            raise RuntimeError("Locales not loaded")
        if isinstance(localized_text, str):
            return localized_text
        translation = self.select_best_translation(localized_text)
        if not translation:
            raise ValueError(
                f"Missing translation for {self.locales} in {localized_text}"
            )
        return translation

    def select_best_translation(self, localized_text: LocalizedText) -> str | None:
        if isinstance(localized_text, str):
            return localized_text
        if not localized_text:
            return None
        if not self.locales:
            raise RuntimeError("Locales not loaded")
        translations = localized_text
        for locale in self.locales:
            translation = translations.get(locale)
            if translation:
                return translation
        translation = next(iter(translations.values()))
        return translation
